//---------------------------------------------------------------------------
#ifndef OpenGlH
#define OpenGlH
//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Forms.hpp>
#include <ExtCtrls.hpp>
#include <DsgnIntf.hpp>

//#define USESGI 1

//#ifdef USESGI
//#include "gl.h"
//#include "glu.h"
//#else
#include <gl\gl.h>
#include <gl\glu.h>
//#endif
#include <stdio.h>
#define VERSION_0 0

class TSceneFile
{
    FILE * file;
    long pos;
    bool ok;
    bool writing;
    int version;
public:
    int Version()
        {
        return version;
        }
    bool OK() {return ok;}
    TSceneFile(const char * fileName, bool write, int version);
    bool Write(void * value, int size);
    bool Read(void * value, int &len, int maxsize);
    ~TSceneFile();

    bool writeString(String value);
    bool readString(int version, String & value, String defval);
    bool writeInt(int value);
    bool readInt(int version, int & value, int defval);
    bool writeFloat(float value);
    bool readFloat(int version, float & value, float defval);
};

#define CUBE_TAG        10
#define SPHERE_TAG      11
#define CYLINDER_TAG    12
#define DISK_TAG        13
#define GROUP_TAG       14
#define QUAD_TAG        15

class PACKAGE TPart : public TObject
{
    class TPartGroup * parent;
    GLfloat position[3];
    GLfloat scale[3];
    GLfloat rotationVector[3];
    GLfloat rotationAngle;
    TNotifyEvent FOnChange;

    void __fastcall SetPX(float Value);
    void __fastcall SetPY(float Value);
    void __fastcall SetPZ(float Value);
    void __fastcall SetRX(float Value);
    void __fastcall SetRY(float Value);
    void __fastcall SetRZ(float Value);
    void __fastcall SetSX(float Value);
    void __fastcall SetSY(float Value);
    void __fastcall SetSZ(float Value);
    void __fastcall SetRotationAngle(float Value);
protected:
    String name;
    virtual String __fastcall GetName();
public:
    void __fastcall CopyTo(TPart * pp);
    void __fastcall SetParent(TPartGroup * pG) {parent = pG;}
    TPartGroup * Parent() {return parent;}
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    static TPart * Load(TSceneFile &);
    TPart();
    __fastcall ~TPart();
    virtual void __fastcall Paint();
    virtual TPart * __fastcall Clone()=0;
    void __fastcall AddChild(TPart *);
    __property float PX = {read=position[0], write=SetPX};
    __property float PY = {read=position[1], write=SetPY};
    __property float PZ = {read=position[2], write=SetPZ};
    __property float RX = {read=rotationVector[0], write=SetRX};
    __property float RY = {read=rotationVector[1], write=SetRY};
    __property float RZ = {read=rotationVector[2], write=SetRZ};
    __property float SX = {read=scale[0], write=SetSX};
    __property float SY = {read=scale[1], write=SetSY};
    __property float SZ = {read=scale[2], write=SetSZ};
    __property float RotationAngle = {read=rotationAngle, write=SetRotationAngle};
    __property String Name = {read=name, write=name};
    __property String FullName = {read=GetName};
    __property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
};

class PACKAGE TPartCube : public TPart
{
    virtual String __fastcall GetName();
public:
    virtual TPart * __fastcall Clone();
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartCube();
    virtual void __fastcall Paint();
};

class PACKAGE TPartSphere : public TPart
{
    virtual String __fastcall GetName();
    int FSlices;
    int FStacks;
    void __fastcall SetSlices(int value);
    void __fastcall SetStacks(int value);
public:
    virtual TPart * __fastcall Clone();
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartSphere();
    virtual void __fastcall Paint();
    __property int Slices = {read = FSlices, write=SetSlices};
    __property int Stacks = {read = FStacks, write=SetStacks};
};

class PACKAGE TPartCylinder : public TPart
{
    virtual String __fastcall GetName();
    int FSlices;
    int FStacks;
    float FBaseRadius;
    float FTopRadius;
    float FHeight;
    void __fastcall SetSlices(int value);
    void __fastcall SetStacks(int value);
    void __fastcall SetBaseRadius(float value);
    void __fastcall SetTopRadius(float value);
    void __fastcall SetHeight(float value);
public:
    virtual TPart * __fastcall Clone();
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartCylinder();
    virtual void __fastcall Paint();
    __property int Slices = {read = FSlices, write=SetSlices};
    __property int Stacks = {read = FStacks, write=SetStacks};
    __property float BaseRadius = {read = FBaseRadius, write=SetTopRadius};
    __property float TopRadius = {read = FTopRadius, write=SetBaseRadius};
    __property float Height = {read = FHeight, write=SetHeight};
};

class PACKAGE TPartDisk : public TPart
{
    int FSlices;
    int FLoops;
    float FInnerRadius;
    float FOuterRadius;
    virtual String __fastcall GetName();
    void __fastcall SetSlices(int value);
    void __fastcall SetLoops(int value);
    void __fastcall SetInnerRadius(float value);
    void __fastcall SetOuterRadius(float value);
public:
    virtual TPart * __fastcall Clone();
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartDisk();
    virtual void __fastcall Paint();
    __property int Slices = {read = FSlices, write=SetSlices};
    __property int Loops = {read = FLoops, write=SetLoops};
    __property float InnerRadius = {read = FInnerRadius, write=SetInnerRadius};
    __property float OuterRadius = {read = FOuterRadius, write=SetOuterRadius};
};

class PACKAGE TPartGroup : public TPart
{
    virtual String __fastcall GetName();
    TList * children;
public:
    virtual TPart * __fastcall Clone();
    void __fastcall Changed(TObject *Sender)
        {
        if (OnChange)
            OnChange(this);
        }
    void __fastcall AddObject(TPart * pEl);
    void __fastcall DeleteObject(TPart * pObj);

    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartGroup();
    __fastcall ~TPartGroup();
    virtual void __fastcall Paint();
__published:
    __property TList * Parts = {read=children};
};

class PACKAGE TPartQuad : public TPart
{
    virtual String __fastcall GetName();
    int FNPoints;
    GLfloat * vertexArr;
    GLfloat * normalArr;
    int FShape;

public:
    virtual TPart * __fastcall Clone();
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    TPartQuad();
    virtual void __fastcall Paint();
    void __fastcall SetNPoints(int value);
    void __fastcall SetX(int n, float value);
    void __fastcall SetY(int n, float value);
    void __fastcall SetZ(int n, float value);
    float __fastcall GetX(int n);
    float __fastcall GetY(int n);
    float __fastcall GetZ(int n);
    void __fastcall SetNormX(int n, float value);
    void __fastcall SetNormY(int n, float value);
    void __fastcall SetNormZ(int n, float value);
    float __fastcall GetNormX(int n);
    float __fastcall GetNormY(int n);
    float __fastcall GetNormZ(int n);
    void __fastcall SetShape(int value);

__published:
    __property int NPoints = {read=FNPoints, write=SetNPoints};
    __property float X[int n] = {read=GetX, write=SetX};
    __property float Y[int n] = {read=GetY, write=SetY};
    __property float Z[int n] = {read=GetZ, write=SetZ};
    __property float NormX[int n] = {read=GetNormX, write=SetNormX};
    __property float NormY[int n] = {read=GetNormY, write=SetNormY};
    __property float NormZ[int n] = {read=GetNormZ, write=SetNormZ};
    __property int Shape = {read=FShape, write=SetShape};
};

class PACKAGE TFloatVector : public TPersistent
{
    GLfloat v[4];
    TNotifyEvent FOnChange;
    float __fastcall GetElement(int n);
    void __fastcall SetElement(int n, float AX);
    void __fastcall Changed();
public:
    TFloatVector()
        {
        v[0]=0;
        v[1]=0;
        v[2]=0;
        v[3]=1;
        }
     void __fastcall Assign(TPersistent * Value);
     const GLfloat * Vector() {return &v[0];}

    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(int ver, TSceneFile & sf, float def0=0.0, float def1=0.0, float def2=0.0, float def3=0.0);

__published:
    __property float Element[int] = {read=GetElement, write=SetElement};
    __property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
};

class PACKAGE T3DVector : public TFloatVector
{
    void __fastcall SetX(float value);
    void __fastcall SetY(float value);
    void __fastcall SetZ(float value);
    void __fastcall SetW(float value);
    float __fastcall GetX();
    float __fastcall GetY();
    float __fastcall GetZ();
    float __fastcall GetW();
public:
    T3DVector(){}
__published:
    __property float X = {read=GetX, write=SetX};
    __property float Y = {read=GetY, write=SetY};
    __property float Z = {read=GetZ, write=SetZ};
    __property float W = {read=GetW, write=SetW};

};

class PACKAGE TColorVector : public TFloatVector
{
    void __fastcall SetR(float value);
    void __fastcall SetG(float value);
    void __fastcall SetB(float value);
    void __fastcall SetA(float value);
    float __fastcall GetR();
    float __fastcall GetG();
    float __fastcall GetB();
    float __fastcall GetA();
public:
    TColorVector(){}
__published:
    __property float R = {read=GetR, write=SetR};
    __property float G = {read=GetG, write=SetG};
    __property float B = {read=GetB, write=SetB};
    __property float A = {read=GetA, write=SetA};

};

class PACKAGE TAngle : public TFloatVector
{
    void __fastcall SetDegrees(float value);
    float __fastcall GetDegrees();
public:
    TAngle(){}
__published:
    __property float Degrees = {read=GetDegrees, write=SetDegrees};
};


//---------------------------------------------------------------------------
class PACKAGE TFloatEdit : public TPanel
{
private:
    TNotifyEvent FOnChange;

    TPanel *topPanel;
    TLabel * xEdit;
    TLabel * yEdit;
    TLabel * zEdit;
    TLabel * wEdit;
    TScrollBar * scroll;
    int FNValues;
    int FDecDigits;
    TStringList * FLabels;
//    double x;
//    double y;
//    double z;
    void __fastcall ReSizeFields(void);
    void __fastcall EditClick(TObject *Sender);
    void __fastcall RefreshCurrent();
    void __fastcall ScrollChange(TObject *Sender);
    void __fastcall RefreshField(int n);
    void __fastcall SetLabelCaptions(TObject *Sender);

    TLabel * CurrEdit;
protected:
    TFloatVector * FValue;
    TFloatVector * FMin;
    TFloatVector * FMax;
    virtual void __fastcall Loaded(void);
    void __fastcall ValueChanged(TObject *Sender);
    void __fastcall SetValue(TFloatVector * Value);
    void __fastcall SetMin(TFloatVector * Value);
    void __fastcall SetMax(TFloatVector * Value);
    void __fastcall SetNValues(int n);
    void __fastcall SetDecDigits(int n);
    void __fastcall SetLabels(TStringList *);

public:
    __fastcall TFloatEdit(TComponent* Owner);
    __fastcall ~TFloatEdit() {delete FValue; delete FMin; delete FMax;delete FLabels;}
__published:
    __property TFloatVector * Value = {read=FValue, write=SetValue};
    __property TFloatVector * Min = {read=FMin, write=SetMin};
    __property TFloatVector * Max = {read=FMax, write=SetMax};
    __property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
    __property int NValues = {read=FNValues, write=SetNValues,default=3};
    __property TStringList * Labels = {read=FLabels, write=SetLabels};
    __property int DecDigits = {read=FDecDigits, write=SetDecDigits,default=2};
};

class PACKAGE TPos3D : public TFloatEdit
{
private:
    T3DVector * __fastcall GetValue() {return (T3DVector *)FValue;}
    void __fastcall SetValue(T3DVector * Value) {TFloatEdit::SetValue(Value);}
    T3DVector * __fastcall GetMin() {return (T3DVector *)FMin;}
    void __fastcall SetMin(T3DVector * Value) {TFloatEdit::SetMin(Value);}
    T3DVector * __fastcall GetMax() {return (T3DVector *)FMax;}
    void __fastcall SetMax(T3DVector * Value) {TFloatEdit::SetMax(Value);}
public:
    __fastcall TPos3D(TComponent* Owner);
__published:
    __property T3DVector * Value = {read=GetValue, write=SetValue};
    __property T3DVector * Min = {read=GetMin, write=SetMin};
    __property T3DVector * Max = {read=GetMax, write=SetMax};
};

class PACKAGE TColorEdit : public TFloatEdit
{
private:
    TColorVector * __fastcall GetValue() {return (TColorVector *)FValue;}
    void __fastcall SetValue(TColorVector * Value) {TFloatEdit::SetValue(Value);}
    TColorVector * __fastcall GetMin() {return (TColorVector *)FMin;}
    void __fastcall SetMin(TColorVector * Value) {TFloatEdit::SetMin(Value);}
    TColorVector * __fastcall GetMax() {return (TColorVector *)FMax;}
    void __fastcall SetMax(TColorVector * Value) {TFloatEdit::SetMax(Value);}
public:
    __fastcall TColorEdit(TComponent* Owner);
__published:
    __property TColorVector * Value = {read=GetValue, write=SetValue};
    __property TColorVector * Min = {read=GetMin, write=SetMin};
    __property TColorVector * Max = {read=GetMax, write=SetMax};
};

class PACKAGE TAngleEdit : public TFloatEdit
{
private:
    TAngle * __fastcall GetValue() {return (TAngle *)FValue;}
    void __fastcall SetValue(TAngle * Value) {TFloatEdit::SetValue(Value);}
    TAngle * __fastcall GetMin() {return (TAngle *)FMin;}
    void __fastcall SetMin(TAngle * Value) {TFloatEdit::SetMin(Value);}
    TAngle * __fastcall GetMax() {return (TAngle *)FMax;}
    void __fastcall SetMax(TAngle * Value) {TFloatEdit::SetMax(Value);}
public:
    __fastcall TAngleEdit(TComponent* Owner);
__published:
    __property TAngle * Value = {read=GetValue, write=SetValue};
    __property TAngle * Min = {read=GetMin, write=SetMin};
    __property TAngle * Max = {read=GetMax, write=SetMax};
};
//---------------------------------------------------------------------------
#define MAXBITMAPS 9

const int MAXLIGHTS = GL_MAX_LIGHTS > 32 ? 32 : GL_MAX_LIGHTS;
class TLuce :  public TPersistent
{
    bool FEnabled;
    TNotifyEvent FOnChange;
    void __fastcall SetEnabled(bool Value)
        {
        FEnabled = Value;
        Changed(this);
        }
    void __fastcall Changed(TObject *Sender)
        {
        if (OnChange)
            OnChange(this);
        }
public:
    virtual bool SaveTo(TSceneFile &);
    virtual bool LoadFrom(TSceneFile &);
    T3DVector * Position;
    TColorVector * Ambient;
    TColorVector * Diffuse;
    TColorVector * Specular;
    void __fastcall Assign(TPersistent * Value);
    TLuce();
    virtual __fastcall ~TLuce();

__published:
    __property TNotifyEvent OnChange = {read=FOnChange, write=FOnChange};
    __property bool Enabled = {read=FEnabled, write=SetEnabled};
};

class PACKAGE TOpenGl : public TWinControl
{
private:
    HGLRC hGLRC;

    TLuce *luci[MAXLIGHTS];
    void __fastcall SetLuci(TObject *Sender);

    TPart * FSuggestedSelection;
    float FFps;
    bool  FAnimation;

    bool  FProjectionOrtho;
    float FProjectionWidth;
    float FProjectionHeight;
    float FProjectionNear;
    float FProjectionFar;
    float FProjectionYAngle;

    T3DVector * FEyePosition;
    T3DVector * FEyeTarget;
    T3DVector * FEyeUp;

    TNotifyEvent FOnListChanged;
    TNotifyEvent FOnFrame;
    TNotifyEvent FOnPaint;
    TNotifyEvent FOnInitRC;
    TNotifyEvent FOnSetProjection;
    TControlCanvas * FCanvas;

    TCanvas * __fastcall GetCanvas() {return FCanvas;}

    LARGE_INTEGER PerfFreq;
    LARGE_INTEGER LastPerfCounter;

    virtual void __fastcall CreateParams(TCreateParams &Params);
    virtual void __fastcall Loaded();
    virtual void __fastcall Paint(void);
    virtual void __fastcall PaintWindow(HDC DC);
    void __fastcall WMPaint(TWMPaint Message);
    void __fastcall WMSize(TWMSize Message);

    void defaultInitRC(void);
    void setupPixelFormat();
    void SetProjection(void);

protected:
    void __fastcall SetAnimation(bool AAnimation);
    void __fastcall SetProjectionOrtho(bool AProjectionOrtho);
    void __fastcall SetProjectionWidth(float AProjectionWidth);
    void __fastcall SetProjectionHeight(float AProjectionHeight);
    void __fastcall SetProjectionNear(float AProjectionNear);
    void __fastcall SetProjectionFar(float AProjectionFar);
    void __fastcall SetProjectionYAngle(float AProjectionYAngle);
    TPicture * __fastcall GetPicture(int n) {return pictures[n];}
    void  __fastcall SetPicture(int n, TPicture * Apicture);
    TPicture * pictures[MAXBITMAPS];
    int    textureIndexes[MAXBITMAPS];
    int __fastcall SetTexture(int index, Graphics::TBitmap * pBitmap);
    TList * parts;
    void __fastcall SetEyePosition(T3DVector * Value);
    void __fastcall SetEyeTarget(T3DVector * Value);
    void __fastcall SetEyeUp(T3DVector * Value);
    void __fastcall SetLight(int n, TLuce * Value);
    TLuce * __fastcall GetLight(int n);
    int __fastcall GetResourceContext() {return (int)hGLRC;}
public:

    bool SaveTo(String fileName);
    bool LoadFrom(String fileName);
    int __fastcall UseTexture(int index);
    __fastcall TOpenGl(TComponent* Owner);
    __fastcall ~TOpenGl();
    void __fastcall Refresh(TObject *Sender);
    TPart * __fastcall AddGroup(TPart * pParent, const char * name = 0);
    TPart * __fastcall AddQuad(TPart * pParent, const char * name = 0);
    TPart * __fastcall AddCube(TPart * pParent, const char * name = 0);
    TPart * __fastcall AddSphere(TPart * pParent, const char * name = 0);
    TPart * __fastcall AddCylinder(TPart * pParent, const char * name = 0);
    TPart * __fastcall AddDisk(TPart * pParent, const char * name = 0);
    void __fastcall DeleteObject(TPart *);
    void __fastcall AddObject(TPart * pParent, TPart * pObj);
    TPart * __fastcall Paste(TPart * pParent, TPart * pp);
__published:
    __property TPicture * Textures[int n] = {read=GetPicture, write=SetPicture};
    __property TCanvas * Canvas = {read=GetCanvas};
    __property TNotifyEvent OnSetProjection = {read=FOnSetProjection, write=FOnSetProjection};
    __property TNotifyEvent OnPaint = {read=FOnPaint, write=FOnPaint};
    __property TNotifyEvent OnFrame = {read=FOnFrame, write=FOnFrame};
    __property TNotifyEvent OnInitRC = {read=FOnInitRC, write=FOnInitRC};
    __property TNotifyEvent OnListChanged = {read=FOnListChanged, write=FOnListChanged};
    __property bool Animation = {read=FAnimation, write=SetAnimation, default=false};
    __property float Fps = {read=FFps};
    __property Visible;
    __property Align;

    __property bool ProjectionOrtho = {read=FProjectionOrtho, write=SetProjectionOrtho, default = false};
    __property float ProjectionWidth = {read=FProjectionWidth, write=SetProjectionWidth, default = 1};
    __property float ProjectionHeight = {read=FProjectionHeight, write=SetProjectionHeight, default=1};
    __property float ProjectionNear = {read=FProjectionNear, write=SetProjectionNear, default=1};
    __property float ProjectionFar = {read=FProjectionFar, write=SetProjectionFar, default=10};
    __property float ProjectionYAngle = {read=FProjectionYAngle, write=SetProjectionYAngle, default=90};
    __property T3DVector * EyePosition = {read=FEyePosition, write=SetEyePosition};
    __property T3DVector * EyeTarget = {read=FEyeTarget, write=SetEyeTarget};
    __property T3DVector * EyeUp = {read=FEyeUp, write=SetEyeUp};
    __property TList * Parts = {read=parts};
    __property TPart * SuggestedSelection = {read=FSuggestedSelection};
    __property TLuce * Lights[int n] = {read=GetLight, write=SetLight};
    __property int ResourceContext = {read=GetResourceContext};
BEGIN_MESSAGE_MAP
  MESSAGE_HANDLER(WM_PAINT, TWMPaint, WMPaint)
  MESSAGE_HANDLER(WM_SIZE, TWMSize, WMSize)
END_MESSAGE_MAP(TComponent)
};
//---------------------------------------------------------------------------
#endif
