//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <stdio.h>

#include "OpenGl.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#include <values.h>

void __fastcall TPart::CopyTo(TPart * pp)
{
    pp->parent=0;
    memcpy(&pp->position, &position, sizeof(float)*3);
    memcpy(&pp->scale, &scale, sizeof(float)*3);
    memcpy(&pp->rotationVector, &rotationVector, sizeof(float)*3);
    pp->rotationAngle = rotationAngle;
    pp->name = name;
}

TPart * __fastcall TPartCube::Clone()
{
    TPart * pp = new TPartCube();
    CopyTo(pp);
    return pp;
}

TPart * __fastcall TPartSphere::Clone()
{
    TPartSphere * pp = new TPartSphere();
    CopyTo(pp);
    pp->FSlices = FSlices;
    pp->FSlices = FStacks;
    return pp;
}

TPart * __fastcall TPartCylinder::Clone()
{
    TPartCylinder * pp = new TPartCylinder();
    CopyTo(pp);
    pp->FSlices = FSlices;
    pp->FStacks = FStacks;
    pp->FBaseRadius = FBaseRadius;
    pp->FTopRadius = FTopRadius;
    pp->FHeight = FHeight;
    return pp;
}

TPart * __fastcall TPartDisk::Clone()
{
    TPartDisk * pp = new TPartDisk();
    CopyTo(pp);
    pp->FSlices = FSlices;
    pp->FLoops = FLoops;
    pp->FInnerRadius = FInnerRadius;
    pp->FOuterRadius = FOuterRadius;
    return pp;
}

TPart * __fastcall TPartGroup::Clone()
{
    TPartGroup * pp = new TPartGroup();
    CopyTo(pp);
    for (int i=0; i<children->Count; i++)
        {
        TPart *pP = (TPart *)children->Items[i];
        pp->AddObject(pP->Clone());
        }
    return pp;
}

TPart * TPart::Load(TSceneFile & sf)
{
    bool good = sf.OK();
    int tag=0;
    good = good && sf.readInt(0, tag, 0);
    TPart * pP=0;
    if (good)
        {
        switch(tag)
            {
            case CUBE_TAG:
                pP = new TPartCube();
                break;
            case SPHERE_TAG:
                pP = new TPartSphere();
                break;
            case CYLINDER_TAG:
                pP = new TPartCylinder();
                break;
            case DISK_TAG:
                pP = new TPartDisk();
                break;
            case GROUP_TAG:
                pP = new TPartGroup();
                break;
            case QUAD_TAG:
                pP = new TPartQuad();
                break;
            }
        }
    if (pP)
        pP->LoadFrom(sf);
    return pP;
}


TPart::TPart()
{
    parent = 0;
    position[0] = 0.0;
    position[1] = 0.0;
    position[2] = 0.0;
    scale[0] = 1.0;
    scale[1] = 1.0;
    scale[2] = 1.0;
    rotationVector[0] = 0.0;
    rotationVector[1] = 0.0;
    rotationVector[2] = 0.0;
    rotationAngle = 0.0;
}

__fastcall TPart::~TPart()
{
}

String __fastcall TPart::GetName()
{
    if (name.Length() > 0)
        return name;
    else
        {
        int i = reinterpret_cast<int>(this);
        char mmm[100];
        sprintf(mmm, "ID:%x", i);
        String s = mmm;
        return s;
        }
}

void __fastcall TPart::Paint()
{
    glTranslatef(position[0],
                 position[1],
                 position[2]);
    glRotatef(rotationAngle,    rotationVector[0],
                                rotationVector[1],
                                rotationVector[2]);
    glScalef(scale[0],
             scale[1],
             scale[2]);
}

void __fastcall TPart::AddChild(TPart *)
{
}

void __fastcall TPart::SetPX(float Value)
{
    position[0] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetPY(float Value)
{
    position[1] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetPZ(float Value)
{
    position[2] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetRX(float Value)
{
    rotationVector[0] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetRY(float Value)
{
    rotationVector[1] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetRZ(float Value)
{
    rotationVector[2] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetSX(float Value)
{
    scale[0] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetSY(float Value)
{
    scale[1] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetSZ(float Value)
{
    scale[2] = Value;
    if (FOnChange)
        FOnChange(this);
}

void __fastcall TPart::SetRotationAngle(float Value)
{
    rotationAngle = Value;
    if (FOnChange)
        FOnChange(this);
}



TPartCube::TPartCube()
{
}

void __fastcall TPartCube::Paint()
{
    TPart::Paint();
    GLubyte frontIndices[] = {4,5,6,7};
    GLubyte rightIndices[] = {1,2,6,5};
    GLubyte bottomIndices[] = {0,1,5,4};
    GLubyte backIndices[] = {0,3,2,1};
    GLubyte leftIndices[] = {0,4,7,3};
    GLubyte topIndices[] = {2,3,7,6};
//   glEnableClientState(GL_VERTEX_ARRAY);
//   glVertexPointer(3, GL_FLOAT, 0, vertices);
    glNormal3f( 0, 0.0, 1.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, frontIndices);
    glNormal3f( 1.0, 0.0, 0.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, rightIndices);
    glNormal3f( 0.0, -1.0, 0.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, bottomIndices);
    glNormal3f( 0.0, 0.0, -1.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, backIndices);
    glNormal3f( -1.0, 0.0, 0.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, leftIndices);
    glNormal3f( 0.0, 1.0, 0.0);
    glDrawElements(GL_QUADS, 4, GL_UNSIGNED_BYTE, topIndices);
}

String __fastcall TPartCube::GetName()
{
    String s = "Cube ";
    s += TPart::GetName();
    return s;
}

TPartSphere::TPartSphere()
{
    FSlices = 20;
    FStacks = 20;
}

void __fastcall TPartSphere::SetSlices(int value)
{
    FSlices=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartSphere::SetStacks(int value)
{
    FStacks = value;
    if (OnChange)
        OnChange(this);
}


void __fastcall TPartSphere::Paint()
{
    TPart::Paint();
    GLUquadricObj *quadric = gluNewQuadric();
    gluQuadricNormals( quadric,
			  GLU_SMOOTH );
    gluQuadricDrawStyle(quadric,
			    GLU_FILL );
    gluQuadricTexture( quadric,
			  GL_TRUE );
    gluSphere( quadric,
		  (GLdouble)1,
		  FSlices,
		  FStacks);
    gluDeleteQuadric(quadric);
}

String __fastcall TPartSphere::GetName()
{
    String s = "Sphere ";
    s += TPart::GetName();
    return s;
}



TPartCylinder::TPartCylinder()
{
    FSlices = 20;
    FStacks = 20;
    FBaseRadius = 1;
    FTopRadius = 1;
    FHeight = 1;

}

void __fastcall TPartCylinder::SetBaseRadius(float value)
{
    FBaseRadius=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartCylinder::SetTopRadius(float value)
{
    FTopRadius=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartCylinder::SetHeight(float value)
{
    FHeight=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartCylinder::SetSlices(int value)
{
    FSlices=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartCylinder::SetStacks(int value)
{
    FStacks = value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartCylinder::Paint()
{
    TPart::Paint();
    GLUquadricObj *quadric = gluNewQuadric();
    gluQuadricNormals( quadric,
			  GLU_SMOOTH );
    gluQuadricDrawStyle(quadric,
			    GLU_FILL );
    gluQuadricTexture( quadric,
			  GL_TRUE );
    gluCylinder( quadric,
		        FBaseRadius,
		        FTopRadius,
		        FHeight,
		        FSlices,
		        FStacks);
    gluDeleteQuadric(quadric);
}

String __fastcall TPartCylinder::GetName()
{
    String s = "Cylinder ";
    s += TPart::GetName();
    return s;
}


TPartDisk::TPartDisk()
{
    FSlices = 20;
    FLoops = 20;
    FInnerRadius = 1;
    FOuterRadius = 2;
}

void __fastcall TPartDisk::Paint()
{
    TPart::Paint();
    GLUquadricObj *quadric = gluNewQuadric();
    gluQuadricNormals( quadric,
			  GLU_SMOOTH );
    gluQuadricDrawStyle(quadric,
			    GLU_FILL );
    gluQuadricTexture( quadric,
			  GL_TRUE );
    gluDisk(quadric,
		FInnerRadius,
		FOuterRadius,
		FSlices,
		FLoops);
    gluDeleteQuadric(quadric);
}

String __fastcall TPartDisk::GetName()
{
    String s = "Disk ";
    s += TPart::GetName();
    return s;
}

void __fastcall TPartDisk::SetInnerRadius(float value)
{
    FInnerRadius=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartDisk::SetOuterRadius(float value)
{
    FOuterRadius=value;
    if (OnChange)
        OnChange(this);
}


void __fastcall TPartDisk::SetSlices(int value)
{
    FSlices=value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartDisk::SetLoops(int value)
{
    FLoops = value;
    if (OnChange)
        OnChange(this);
}


//////////////////////////////////// files

bool TPart::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeString(name);
    good = good && sf.writeFloat(position[0]);
    good = good && sf.writeFloat(position[1]);
    good = good && sf.writeFloat(position[2]);
    good = good && sf.writeFloat(scale[0]);
    good = good && sf.writeFloat(scale[1]);
    good = good && sf.writeFloat(scale[2]);
    good = good && sf.writeFloat(rotationVector[0]);
    good = good && sf.writeFloat(rotationVector[1]);
    good = good && sf.writeFloat(rotationVector[2]);
    good = good && sf.writeFloat(rotationAngle);
    return good;
}

bool TPart::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.readString(VERSION_0, name, "");
    good = good && sf.readFloat(VERSION_0, position[0], 0.0);
    good = good && sf.readFloat(VERSION_0, position[1], 0.0);
    good = good && sf.readFloat(VERSION_0, position[2], 0.0);
    good = good && sf.readFloat(VERSION_0, scale[0], 1.0);
    good = good && sf.readFloat(VERSION_0, scale[1], 1.0);
    good = good && sf.readFloat(VERSION_0, scale[2], 1.0);
    good = good && sf.readFloat(VERSION_0, rotationVector[0], 0.0);
    good = good && sf.readFloat(VERSION_0, rotationVector[1], 0.0);
    good = good && sf.readFloat(VERSION_0, rotationVector[2], 0.0);
    good = good && sf.readFloat(VERSION_0, rotationAngle, 0.0);
    return good;
}

bool TPartCube::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(CUBE_TAG);
    good = good && TPart::SaveTo(sf);
    return good;
}

bool TPartSphere::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(SPHERE_TAG);
    good = good && TPart::SaveTo(sf);
    good = good && sf.writeInt(FSlices);
    good = good && sf.writeInt(FStacks);
    return good;
}

bool TPartCylinder::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(CYLINDER_TAG);
    good = good && TPart::SaveTo(sf);
    good = good && sf.writeInt(FSlices);
    good = good && sf.writeInt(FStacks);
    good = good && sf.writeFloat(FBaseRadius);
    good = good && sf.writeFloat(FTopRadius);
    good = good && sf.writeFloat(FHeight);
    return good;
}

bool TPartDisk::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(DISK_TAG);
    good = good && TPart::SaveTo(sf);
    good = good && sf.writeInt(FSlices);
    good = good && sf.writeInt(FLoops);
    good = good && sf.writeFloat(FInnerRadius);
    good = good && sf.writeFloat(FOuterRadius);
    return good;
}


///
bool TPartCube::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);
    return good;
}

bool TPartSphere::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);
    good = good && sf.readInt(VERSION_0, FSlices, 20);
    good = good && sf.readInt(VERSION_0, FStacks, 20);
    return good;
}

bool TPartCylinder::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);
    good = good && sf.readInt(VERSION_0, FSlices, 20);
    good = good && sf.readInt(VERSION_0, FStacks, 20);
    good = good && sf.readFloat(VERSION_0, FBaseRadius, 1);
    good = good && sf.readFloat(VERSION_0, FTopRadius, 1);
    good = good && sf.readFloat(VERSION_0, FHeight, 1);
    return good;
}

bool TPartDisk::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);
    good = good && sf.readInt(VERSION_0, FSlices, 20);
    good = good && sf.readInt(VERSION_0, FLoops, 20);
    good = good && sf.readFloat(VERSION_0, FInnerRadius, 1);
    good = good && sf.readFloat(VERSION_0, FOuterRadius, 2);
    return good;
}

TPartGroup::TPartGroup()
{
    children = new TList();
}

void __fastcall TPartGroup::AddObject(TPart * pEl)
{
  children->Add(pEl);
  pEl->SetParent(this);
  if (OnChange)
    OnChange(this);
}

__fastcall TPartGroup::~TPartGroup()
{
   // cancella i figli !!!!!!!
   delete children;
}

void __fastcall TPartGroup::DeleteObject(TPart * pObj)
{
  int idx = children->IndexOf(pObj);
  if (idx >= 0)
    children->Delete(idx);
  if (OnChange)
    OnChange(this);
}

void __fastcall TPartGroup::Paint()
{
    TPart::Paint();
    for (int i=0; i<children->Count; i++)
        {
        glPushMatrix();
        TPart *pP = (TPart *)children->Items[i];
        pP->Paint();
        glPopMatrix();
        }
}

String __fastcall TPartGroup::GetName()
{
    String s = "Group ";
    s += TPart::GetName();
    return s;
}


bool TPartGroup::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(GROUP_TAG);
    good = good && TPart::SaveTo(sf);
    sf.writeInt(children->Count);

    for (int i=0; i<children->Count; i++)
        {
        TPart *pP = (TPart *)children->Items[i];
        pP->SaveTo(sf);
        }

    return good;
}

bool TPartGroup::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);

    int c=0;
    children->Clear();
    sf.readInt(VERSION_0, c, 0);
    for (int i=0; i<c; i++)
        {
        TPart * pP = TPart::Load(sf);
        if (pP)
            {
            pP->OnChange = Changed;
            pP->SetParent(this);
            children->Add(pP);
            }
        }

    return good;
}

// QUAD
String __fastcall TPartQuad::GetName()
{
    String s = "Quad ";
    s += TPart::GetName();
    return s;
}

TPart * __fastcall TPartQuad::Clone()
{
    TPartQuad * pp = new TPartQuad();
    CopyTo(pp);
    pp->Shape = Shape;
    pp->NPoints = NPoints;
    for (int i=0; i<FNPoints; i++)
        {
        pp->vertexArr[i*3] = vertexArr[i*3];
        pp->vertexArr[i*3+1] = vertexArr[i*3+1];
        pp->vertexArr[i*3+2] = vertexArr[i*3+2];
        pp->normalArr[i*3] = normalArr[i*3];
        pp->normalArr[i*3+1] = normalArr[i*3+1];
        pp->normalArr[i*3+2] = normalArr[i*3+2];
        }
    return pp;
}

bool TPartQuad::SaveTo(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && sf.writeInt(QUAD_TAG);
    good = good && TPart::SaveTo(sf);
    good = good && sf.writeInt(FShape);
    good = good && sf.writeInt(FNPoints);
    for (int i=0; i<FNPoints*3; i++)
        good = good && sf.writeFloat(vertexArr[i]);
    for (int i=0; i<FNPoints*3; i++)
        good = good && sf.writeFloat(normalArr[i]);
    return good;
}

bool TPartQuad::LoadFrom(TSceneFile & sf)
{
    bool good = sf.OK();
    good = good && TPart::LoadFrom(sf);
    good = good && sf.readInt(VERSION_0, FShape, GL_QUADS);
    int tmp;
    good = good && sf.readInt(VERSION_0, tmp, 4);
    NPoints = tmp;
    for (int i=0; i<NPoints*3; i++)
        good = good && sf.readFloat(VERSION_0, vertexArr[i], 0.0);
    for (int i=0; i<NPoints*3; i++)
        good = good && sf.readFloat(VERSION_0, normalArr[i], 0.0);
    return good;
}

void __fastcall TPartQuad::SetNPoints(int value)
{
    GLfloat * tmpArr = new GLfloat[value*3];
    GLfloat * tmpNArr = new GLfloat[value*3];
    if (vertexArr)
        {
        int i=0;
        int n = (value<FNPoints)? value : FNPoints;
        for (; i<n ;i++)
            {
            tmpArr[i*3] = vertexArr[i*3];
            tmpArr[i*3+1] = vertexArr[i*3+1];
            tmpArr[i*3+2] = vertexArr[i*3+2];
            tmpNArr[i*3] = normalArr[i*3];
            tmpNArr[i*3+1] = normalArr[i*3+1];
            tmpNArr[i*3+2] = normalArr[i*3+2];
            }
        for (; i<value;i++)
            {
            tmpArr[i*3] = 0.0;
            tmpArr[i*3+1] = 0.0;
            tmpArr[i*3+2] = 0.0;
            tmpNArr[i*3] = 0.0;
            tmpNArr[i*3+1] = 0.0;
            tmpNArr[i*3+2] = 0.0;
            }
        delete vertexArr;
        delete normalArr;
        }
    FNPoints = value;
    vertexArr = tmpArr;
    normalArr = tmpNArr;
    if (OnChange)
        OnChange(this);
}

TPartQuad::TPartQuad()
{
    FShape = GL_QUADS;
    vertexArr = 0;
    normalArr = 0;
    NPoints = 4;
    vertexArr[0] = -1;
    vertexArr[1] = -1;
    vertexArr[2] = 0;
    vertexArr[3] = 1;
    vertexArr[4] = -1;
    vertexArr[5] = 0;
    vertexArr[6] = 1;
    vertexArr[7] = 1;
    vertexArr[8] = 0;
    vertexArr[9] = -1;
    vertexArr[10] = 1;
    vertexArr[11] = 0;
    normalArr[0] = 0;
    normalArr[1] = 0;
    normalArr[2] = 1;
    normalArr[3] = 0;
    normalArr[4] = 0;
    normalArr[5] = 1;
    normalArr[6] = 0;
    normalArr[7] = 0;
    normalArr[8] = 1;
    normalArr[9] = 0;
    normalArr[10] = 0;
    normalArr[11] = 1;
}

void __fastcall TPartQuad::Paint()
{
    TPart::Paint();
    glBegin(Shape);
    for (int i=0; i<FNPoints; i++)
        {
        glVertex3fv(&vertexArr[i*3]);
        glNormal3fv(&normalArr[i*3]);
        }
    glEnd();
}

void __fastcall TPartQuad::SetX(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        vertexArr[n*3] = value;
        if (OnChange)
            OnChange(this);
        }
}

void __fastcall TPartQuad::SetY(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        vertexArr[n*3 + 1] = value;
        if (OnChange)
            OnChange(this);
        }
}

void __fastcall TPartQuad::SetZ(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        vertexArr[n*3 + 2] = value;
        if (OnChange)
            OnChange(this);
        }
}

float __fastcall TPartQuad::GetX(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return vertexArr[n*3];
        }
    else
        return 0;
}

float __fastcall TPartQuad::GetY(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return vertexArr[n*3 + 1];
        }
    else
        return 0;
}

float __fastcall TPartQuad::GetZ(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return vertexArr[n*3 + 2];
        }
    else
        return 0;
}

void __fastcall TPartQuad::SetShape(int value)
{
    FShape = value;
    if (OnChange)
        OnChange(this);
}

void __fastcall TPartQuad::SetNormX(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        normalArr[n*3] = value;
        if (OnChange)
            OnChange(this);
        }
}

void __fastcall TPartQuad::SetNormY(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        normalArr[n*3+1] = value;
        if (OnChange)
            OnChange(this);
        }
}

void __fastcall TPartQuad::SetNormZ(int n, float value)
{
    if (n>=0 && n<FNPoints)
        {
        normalArr[n*3+2] = value;
        if (OnChange)
            OnChange(this);
        }
}

float __fastcall TPartQuad::GetNormX(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return normalArr[n*3];
        }
    else
        return 0;
}

float __fastcall TPartQuad::GetNormY(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return normalArr[n*3 + 1];
        }
    else
        return 0;
}

float __fastcall TPartQuad::GetNormZ(int n)
{
    if (n>=0 && n<FNPoints)
        {
        return normalArr[n*3 + 2];
        }
    else
        return 0;
}

